#!/usr/bin/env python
# -*- coding: utf8 -*-
""" risk management rules """

class SingleMarketsException(Exception):
  """ Single Markets - A maximum of 4 Units per market. """


class CloselyCorrelatedMarketsException(Exception):
  """ Closely Correlated Markets - A maximum of 6 Units in one particular direction. """


class LooselyCorrelatedMarketsException(Exception):
  """ Loosely Correlated Markets - A maximum of 10 Units in one particular direction. """


class SingleDirectionException(Exception):
  """ Single Direction - A maximum of 12 Units in one direction, long or short.  """


# #!/usr/bin/env python
# # -*- coding: UTF8 -*-
# """https://oxfordstrat.com/coasdfASD32/uploads/2016/01/turtle-rules.pdf

# Exits
# The Turtles used breakout based exits for profitable positions.

# There is an old saying: "you can never go broke taking a profit." The 
# Turtles would not agree with this statement. Getting out of winning positions
# too early, i.e. "taking a profit" too early, is one of the most common mistakes
# when trading trend following systems.

# Prices never go straight up; therefore it is necessary to let the prices go against you if
# you are going to ride a trend. Early in a trend this can often mean watching decent
# profits of 10% to 30% fade to a small loss. In the middle of a trend, it might mean
# watching a profit of 80% to 100% drop by 30% to 40%. The temptation to lighten the 
# position to "lock in profits" can be very great.

# The Turtles knew that where you took a profit could make the difference between
# winning and losing.

# The Turtle System enters on breakouts. Most breakouts do not result in trends. This
# means that most of the trades that the Turtles made resulted in losses. If the winning
# trades did not earn enough on average to offset these losses, the Turtles would have
# lost money. Every profitable trading system has a different optimal exit point.

# Consider the Turtle System; if you exit winning positions at a 1 N profit while you
# exited losing positions at a 2 N loss you would need twice as many winners to offset
# the losses from the losing trades.

# There is a complex relationship between the components of a trading system. This
# means that you can't consider the proper exit for a profitable position without
# considering the entry, money management and other factors.

# The proper exit for winning positions is one of the most important aspects of trading,
# and the least appreciated. Yet it can make the difference between winning and losing.

# Turtle Exits

# The System 1 exit was a 10 day low for long positions and a 10 day high for short
# positions. All the Units in the position would be exited if the price went against the
# position for a 10 day breakout.

# The System 2 exit was a 20 day low for long positions and a 20 day high for short
# positions. All the Units in the position would be exited if the price went against the
# position for a 20 day breakout.

# As with entries, the Turtles did not typically place exit stop orders, but instead watched
# the price during the day, and started to phone in exit orders as soon as the price traded
# through the exit breakout price.
# """

# # n
# # breakout
# # long or short: bool

# def getexit(n, breakout, type: bool):
#   """Return a decision on whether to exit or not.

#   Args:
#     n: 'N' of the breakout.
#     breakout: The breakout price, i.e. the first unit entered.
#     units: How many units in that same position there are.
#   """
#   return Exit(n=n, breakout=breakout, type=type)


# class Exit:
#   def __init__(self, n, breakout, type: bool):

#     self.exit_signal = self.signal()
  
  

#   def signal(self) -> bool:
#     """Return a boolean determing whether to exit or not.

#     Args:
      

#     Returns:
#       True to exit, False to not. 
#     """
#     # First, get the  


#!/usr/bin/env python
# -*- coding: utf8 -*-
""" adding units """
from typing import Optional

def addunits(breakout: float, n: float, unit_list: Optional[int] = 4):
  """ adding units to existing positions """
  units = [breakout]
  for i in range(unit_list-1):
    units.append(round((units[i] + 0.5 * n), 4))

  return units


#!/usr/bin/env python
# -*- coding: utf8 -*-
""" the underlying volatility of an asset """
import warnings
import datetime
import pandas as pd
from typing import Optional

from turtle_trading.dataframe_loader import DataFrameLoader

""" ignore Pandas Future Warning and SettingWithCopyWarnings """
warnings.simplefilter(action='ignore', category=FutureWarning)
pd.set_option('mode.chained_assignment', None)


def getn(dataframe: DataFrameLoader, date: Optional[datetime.date] = None):
  """ shortcut function for class: N  """
  dataframe.reset()
  return N(dataframe, date).n


class N:
  """ this class represents the process for calculating `N` """
  def __init__(self, dataframe: DataFrameLoader, date: Optional[datetime.date] = None):
    self.dataframe = dataframe
    self.date = date

    self.edit_dataframe() # converted to dataframe
    self.append_new_columns()
    self.pdn = self.get_pdn()
    self.n = self.get_n()


  def edit_dataframe(self) -> pd.DataFrame:
    """ edit dataframe """
    self.dataframe.edit_columns(['low', 'high', 'close'])
    self.dataframe.reverse()

    if not self.date:
      self.dataframe = self.dataframe.dataframe.head(20)
      return self.dataframe
    
    self.dataframe = self.dataframe.dataframe[self.date:].head(20)
    return self.dataframe


  def append_new_columns(self) -> pd.DataFrame:
    """ create previous close and true range columns """
    self.dataframe['previous_close'] = self.dataframe['close'].shift(-1)
    self.dataframe['true_range'] = 0
  
    for index, row in enumerate(self.dataframe.iterrows()):
      
      # Set the last 'previous_close' to 0, otherwise will raise NoneType error
      if isinstance(row[1]['previous_close'], type(None)):
        row[1]['previous_close'] = 0

      maximum = max(
        row[1]['high'] - row[1]['low'], row[1]['high'] - row[1]['previous_close'], row[1]['previous_close'] - row[1]['low']
      ) 
      self.dataframe.iloc[index, self.dataframe.columns.get_loc('true_range')] = maximum


  def get_pdn(self) -> float:
    """ calculate the previous day's n """
    return sum(self.dataframe['true_range'][1:]) / 20
  

  def get_n(self) -> float:
    """ calculate n """
    return round(((19 * self.pdn + self.dataframe['true_range'][-1]) / 20), 4)


from .system import getsignal, EntrySignal
from .breakouts.breakout import getbreakouts, check_if_breakout, Breakout
from .breakouts.last_breakout import get_last_breakout, get_last_breakout_profitability
from .add_units import addunits


#!/usr/bin/env python
# -*- coding: utf8 -*-
""" last breakouts and profitability """
import datetime
from typing import Tuple, Optional

from turtle_trading.entries.breakouts.breakout import getbreakouts, check_if_breakout
from turtle_trading.position_sizing.algorithms.get_n import getn
from turtle_trading.dataframe_loader import DataFrameLoader


def get_last_breakout(dataframe: DataFrameLoader, days: int, skip: Optional[int] = 1) -> Tuple[datetime.date, bool]:
  """ get the last breakout date and direction """
  dataframe.reset()
  dataframe.reverse()

  for index, row in enumerate(dataframe.dataframe.iterrows()):

    date = dataframe.dataframe.iloc[index].name.date()
    breakout_tuple = getbreakouts(dataframe, days, date, True)
    
    if check_if_breakout(breakout_tuple[0], breakout_tuple[1], breakout_tuple[2], breakout_tuple[3]) and index+1 > skip:
      return (date, index, True if breakout_tuple[2] >= breakout_tuple[0] else False)
      
    
def get_last_breakout_profitability(dataframe: DataFrameLoader, stand_devs: float, breakout_tuple: Tuple[datetime.date, int, bool]) -> bool:
  """ get the last breakout profitability """
  dataframe.reset()
  dataframe.reverse()
  n = getn(dataframe, breakout_tuple[0])
  price = dataframe.dataframe.iloc[breakout_tuple[1]].close
 
  for index, row in enumerate(dataframe.dataframe.iloc[:breakout_tuple[1]].iterrows()):

    if index == 10: # before a profitable 10-day exit
      break

    if breakout_tuple[2]:
      if row[1]['high'] - price >= n * stand_devs:
        return False
      
    if not breakout_tuple[2]:
      if row[1]['low'] - price <= n * stand_devs:
        return False
    
  return True


#!/usr/bin/env python
# -*- coding: utf8 -*-
""" breakouts """
import datetime
import pandas as pd
from typing import Tuple, Optional

from turtle_trading.dataframe_loader import DataFrameLoader


def getbreakouts(dataframe: DataFrameLoader, days: int, date: Optional[datetime.date] = None, include_current_extrema: Optional[bool] = False) -> Tuple[float]:
  """ shortcut function for class: Breakout """
  dataframe.reset()
  return Breakout(dataframe, days, date, include_current_extrema).breakouts


def check_if_breakout(high: float, low: float, pos_high: float, pos_low: float) -> bool:
  """ shortcut function for checking a breakout """
  if pos_high >= high or pos_low <= low:
    return True


class Breakout:
  """ this class represents finding breakout numbers """
  def __init__(self, dataframe: DataFrameLoader, days: int, date: Optional[datetime.date] = None, include_current_extrema: Optional[bool] = False):
    self.dataframe = dataframe
    self.days = days
    self.date = date
    self.include_current_extrema = include_current_extrema

    self.edit_dataframe()
    self.breakouts = self.find_extrema()

  def edit_dataframe(self) -> pd.DataFrame:
    """ get dataframe """
    self.dataframe.edit_columns(['low', 'high'])
    self.dataframe.reverse()

    if not self.date:
      self.dataframe = self.dataframe.dataframe.head(self.days)
      return self.dataframe
    
    self.dataframe = self.dataframe.dataframe.loc[self.date:].head(self.days)
    return self.dataframe
  
  def find_extrema(self) -> Tuple[float]:
    """ find the extrema """
    maximum, minimum = self.dataframe['high'].max(), self.dataframe['low'].min()
    if self.include_current_extrema:
      pos_high, pos_low = self.dataframe.iloc[0]['high'], self.dataframe.iloc[0]['low']
      return (maximum, minimum, pos_high, pos_low)

    return (maximum, minimum)


#!/usr/bin/env python
# -*- coding: utf8 -*-
""" entry systems """
from typing import Union, Literal

from turtle_trading.entries.breakouts.breakout import getbreakouts, check_if_breakout
from turtle_trading.entries.breakouts.last_breakout import get_last_breakout, get_last_breakout_profitability
from turtle_trading.dataframe_loader import DataFrameLoader


def getsignal(dataframe: DataFrameLoader, system: Union[Literal[1], Literal[2]]) -> bool:
  """ shortcut function for class: EntrySignal """
  dataframe.reset()
  return EntrySignal(dataframe, system).signal


class EntrySignal:
  """ this class represents an entry signal, true for a valid entry, false for invalid """
  def __init__(self, dataframe: DataFrameLoader, system: Union[Literal[1], Literal[2]]):
    self.dataframe = dataframe
    self.system = system

    self.signal = self.get_signal()

  def get_signal(self) -> bool:
    """ get the entry signal """
    if self.system == 2:
      breakout_tuple = getbreakouts(self.dataframe, 55, include_current_extrema=True)

      if check_if_breakout(breakout_tuple[0], breakout_tuple[1], breakout_tuple[2], breakout_tuple[3]):
        return True
      else:
        return False
      
    if self.system == 1:

      breakout_tuple = getbreakouts(self.dataframe, 20, include_current_extrema=True)
      self.dataframe.reset()

      if check_if_breakout(breakout_tuple[0], breakout_tuple[1], breakout_tuple[2], breakout_tuple[3]):
        last_breakout = get_last_breakout(self.dataframe, 20)
        self.dataframe.reset()
        if get_last_breakout_profitability(self.dataframe, 2, last_breakout):
          return False
        
        else: return True


#!/usr/bin/env python
# -*- coding: utf8 -*-
""" volatility adjusted position units """
import datetime
from typing import Optional

from turtle_trading.position_sizing.algorithms.get_n import getn
from turtle_trading.dataframe_loader import DataFrameLoader


def getunitsize(dataframe: DataFrameLoader, account: float, n: Optional[float] = None, date: Optional[datetime.date] = None):
  """ shortcut function for class: Unit """
  dataframe.reset()
  return Unit(dataframe, account, n, date).unit


class Unit:
  """ this class represents the process for calculating unit size """
  def __init__(self, dataframe: DataFrameLoader, account: float, n: Optional[float] = None, date: Optional[datetime.date] = None):
    self.dataframe = dataframe
    self.account = account
    self.n = n
    self.date = date

    if isinstance(n, type(None)) and isinstance(date, type(None)):
      raise AttributeError(f"Set either `n` or `date`.")
    
    if isinstance(n, float) and isinstance(date, datetime.date):
      raise AttributeError(f"Set only one of `n` or `date`.")

    self.dollar_volatility_adjustment()
    self.unit = self.get_unit_size()


  def dollar_volatility_adjustment(self) -> float: 
    """ calculating the dollar volatility of the asset """
    if self.date:

      price = self.dataframe.start_at(self.date).iloc[0]['close']
      n = getn(self.dataframe, self.date)
      self.dollar_volatility = n * price
      return self.dollar_volatility

    if not self.date:

      price = self.dataframe.live_price
      self.dollar_volatility = self.n * price
      return self.dollar_volatility
  

  def get_unit_size(self) -> float:
    """ calculating unit size  """
    return round(((0.01 * self.account) / self.dollar_volatility), 4)


#!/usr/bin/env python
# -*- coding: utf8 -*-
""" initialize a dataframe """
import warnings
import datetime

from yahoo_fin.stock_info import get_data

""" ignore Pandas Future Warning and SettingWithCopyWarnings """
warnings.simplefilter(action='ignore', category=FutureWarning)


class DataFrameLoader:
  """ this class represents the yahoo_fin pd.DataFrame """
  def __init__(self, ticker: str):
    self.ticker = ticker.upper()
    self.dataframe = self.__initalize()

    self.base_dataframe = self.dataframe

  def __initalize(self):
    """ initialize dataframe """
    dataframe = get_data(self.ticker, interval='1d')
    return dataframe
  
  def edit_columns(self, columns: list[str]):
    """ get specific columns """
    self.dataframe = self.dataframe[columns]
    return self.dataframe
  
  def reverse(self):
    """ reverse dataframe """
    self.dataframe = self.dataframe.iloc[::-1]
    return self.dataframe
  
  def start_at(self, date: datetime.date):
    """ start dataframe at a location """
    self.dataframe = self.dataframe[date:]
    return self.dataframe
  
  @property
  def live_price(self):
    """ get live price """
    if str(self.dataframe.iloc[0].name).split("-")[0] != str(datetime.datetime.now().year): 
      # will work if ticker became public before current year, else will not work
      self.reverse()

    return self.dataframe.iloc[0].close
  
  def reset(self):
    """ reset dataframe """
    self.dataframe = self.base_dataframe
    return self.dataframe
  
  def get_price(self, date: datetime.date): 
    """ get price based on date """
    if str(self.dataframe.iloc[0].name).split("-")[0] != str(datetime.datetime.now().year):
      # will work if ticker became public before current year, else will not work
      self.reverse()
    
    return self.start_at(date).iloc[0]['close']


#!/usr/bin/env
# -*- coding: UTF8 -*-
""" the _config module holds package-wide configurables / settings """


# Warning Class
class UnitsWarning(Warning):
  """ A warning for any of the Markets Exceptions. """


futures = [
  # CBOT
  "", # 30 Year U.S. Treasury Bond
  "ZN=F", # 10 Year U.S. Treasury Note

  # CSCE
  "KC=F", # Coffee
  "CC=F", # Cocoa
  "SB=F", # Sugar
  "CT=F", # Cotton

  # CME
  "6S=F", # Swiss Franc
  "", # Deutschmark
  "", # British Pound
  "", # French Franc
  "6J=F", # Japanese Yen
  "6C=F", # Canadian Dollar
  "", # S&P 500 Stock Index
  "", # Eurodollar
  "", # 90 Day U.S. Treasury Bill

  # Comex
  "GC=F", # Gold
  "SI=F", # Silver
  "HG=F", # Copper

  # NYMEX
  "CL=F", # Crude Oil
  "HO=F", # Heating Oil
  "RB=F" # Unleaded Gas
]


#!/usr/bin/env python
# -*- coding: utf8 -*-
""" stop placement """
from itertools import cycle
from typing import Literal, Union


def getstops(unit_list: list[float], n: float, stop_system: Union[Literal['regular'], Literal['whipsaw']] = "regular"):
  """ shortcut function for functions: get_whipsaw_stops and get_stops """
  if stop_system == 'regular':
    return get_stops(unit_list, n)
  
  if stop_system == 'whipsaw':
    return get_whipsaw_stops(unit_list, n)


def get_stops(unit_list: list[float], n: float):
  """ regular stop strategy """
  two_n = 2 * n
  half_n = .5 * n
  third_n = 1.5 * n

  diff_between_units = [round(elem, 2) for elem in [j-i for i, j in zip(unit_list[:-1], unit_list[1:])]]  # find difference between units
  
  if all(i == diff_between_units[0] for i in diff_between_units):
    return [round(unit_list[-1] - two_n, 2) for i in range(len(unit_list))]
  
  if not all(i == diff_between_units[0] for i in diff_between_units):

    licycle = cycle(reversed(unit_list))
    nextelem = next(licycle)
    stops = []
    met = 0

    for index, unit in enumerate(reversed(unit_list)):
      
      if index+1 == len(unit_list): 
        stops.append(stops[-1]) 
        break

      thiselem, nextelem = nextelem, next(licycle)

      if round(thiselem - nextelem, 2) != half_n:
        stops.append(round(thiselem - two_n, 2))

      elif round(thiselem - nextelem, 2) == half_n and met == 0:
        met = met+1
        stops.append(round(thiselem - third_n, 2))
        
      elif round(thiselem - nextelem, 2) == half_n and met != 0:
        stops.append(stops[-1])

        
    return stops[::-1]
  

def get_whipsaw_stops(unit_list: list[float], n: float):
  """ whipsaw stop strategy """
  half_n = .5 * n
  return [round(unit - half_n, 2) for unit in unit_list]


#!/usr/bin/env python
# -*- coding: utf8 -*-
"""  """
import datetime
import pandas as pd
from typing import Union, Literal

from turtle_trading.dataframe_loader import DataFrameLoader


def price_against_position(dataframe: DataFrameLoader, day: int, date_entered: datetime.date):
  """  """
  dataframe.reset()

  dataframe.start_at(date_entered)

  for index, row in enumerate(dataframe.dataframe.iterrows()):
    if index+1 == day:
      return False
    